// ImageSearchDLL.cpp : Defines the entry point for the DLL application.
// Author(s) .........: Dao Van Trong - TRONG.PRO

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <olectl.h>         // For OleLoadPicture(): loads picture objects from streams.
#include <gdiplus.h>        // GDI+ support for advanced image functions.
#include <winuser.h>        // Provides functions such as GetSystemMetrics().
#include <malloc.h>
#include <shellapi.h>       // Provides Shell functions, such as ExtractIcon().
#include <ctype.h>
#include <tchar.h>
#include <math.h>           // for fabsf(): absolute value for floats.
#include <format>          // For formatted string output (C++20).
#include <string>          // For std::string.
#include <string_view>     // For std::string_view.
#include <vector>          // For std::vector container.
#include <cstring>         // For C string functions (memset, strcpy_s, etc.).
#include <cstdint>         // For fixed-width integer types.

#ifdef _MANAGED
#pragma managed(push, off)
#endif

//-------------------------------------------------------------------------------------
// Macro & Helper Definitions
//-------------------------------------------------------------------------------------
#define CLR_DEFAULT 0x808080   // Default background color used in some operations.
#define CLR_NONE    0xFFFFFFFF  // Special color value indicating "no color" or transparency.

//------------------------------------------------------------------------------
// Inline function to convert a color from its BGR stored value to RGB format.
//------------------------------------------------------------------------------
inline COLORREF rgb_to_bgr(DWORD aRGB)
{
    return RGB(GetBValue(aRGB), GetGValue(aRGB), GetRValue(aRGB));
}

//------------------------------------------------------------------------------
// Inline function to convert a multibyte (char*) string to a wide-character (wchar_t*)
// string. This is required by several Windows APIs that expect Unicode strings.
//------------------------------------------------------------------------------
inline int ToWideCharFunc(const char* source, wchar_t* dest, int dest_size_in_wchars)
{
    return MultiByteToWideChar(CP_ACP, 0, source, -1, dest, dest_size_in_wchars);
}

//-------------------------------------------------------------------------------------
// GetErrorMessage: Maps a numeric error code to an error message string.
//-------------------------------------------------------------------------------------
static const char* GetErrorMessage(int iErrorCode)
{
    switch (iErrorCode)
    {
    case -1: return "Invalid path or file does not exist";  // File not found.
    case -2: return "Failed to load image";                   // Image loading error.
    case -3: return "Failed to get display context";          // Cannot get display DC.
    case -4: return "Failed to retrieve image bits";          // Pixel extraction failed.
    case -5: return "Failed to create compatible DC for screen capture";  // Creating memory DC failed.
    case -6: return "Failed to create screen compatible bitmap";            // Bitmap creation error.
    case -7: return "Failed to select object into DC";                    // GDI object selection error.
    case -8: return "BitBlt operation failed";                             // Screen copy (BitBlt) error.
    case -9: return "Failed to get bits from screen capture";             // Getting pixel bits failed.
    case -10: return "Failed to get bits from scaled image";               // Failed to extract bits from scaled image.
    default: return "Unknown error";  // Fallback for unhandled error codes.
    }
}

//-------------------------------------------------------------------------------------
// LoadPicture: Loads an image (or icon) from a file.
//-------------------------------------------------------------------------------------
// Parameters:
//   sFileImage             - File path (and possibly additional info) of the image/icon.
//   iWidth, iHeight       - On input: desired dimensions; On output: actual dimensions.
//   iTypeImage            - Output variable to indicate the image type (bitmap, icon, etc.).
//   iIconNumber           - If the file contains multiple icons, select which one.
//   bUseGDIPlusIfAvailable- Enables use of GDI+ for loading if available.
// Returns: A handle (HBITMAP) to the loaded image or nullptr on failure.
HBITMAP LoadPicture(char* sFileImage, int iWidth, int iHeight, int& iTypeImage, int iIconNumber, bool bUseGDIPlusIfAvailable)
{
    HBITMAP _hBitmap = nullptr;
    iTypeImage = -1; // Initially, the image type is undetermined.

    // Check if the file specification is valid.
    if (!sFileImage || sFileImage[0] == '\0')
        return nullptr;

    // Negative icon numbers are forbidden.
    if (iIconNumber < 0)
        iIconNumber = 0;

    // Determine the file extension by finding the last dot.
    char* sFile_EXT = strrchr(sFileImage, '.');
    if (sFile_EXT)
        ++sFile_EXT; // Point to the extension (skip the dot).

    // Determine if we should extract an icon (from exe, dll, etc.) or load as image.
    bool ExtractIcon_was_used = iIconNumber > 1 || (sFile_EXT && (
        !_stricmp(sFile_EXT, "exe") ||
        !_stricmp(sFile_EXT, "dll") ||
        !_stricmp(sFile_EXT, "icl") ||
        !_stricmp(sFile_EXT, "cpl") ||
        !_stricmp(sFile_EXT, "scr")));

    if (ExtractIcon_was_used)
    {
        iTypeImage = IMAGE_ICON;
        // Use ExtractIcon() to get the icon handle.
        _hBitmap = reinterpret_cast<HBITMAP>(
            ExtractIcon(GetModuleHandle(nullptr), sFileImage, (iIconNumber > 0 ? iIconNumber - 1 : 0)));
        // If the returned handle is less than 2, it is invalid.
        if (reinterpret_cast<std::uintptr_t>(_hBitmap) < 2)
            return nullptr;
    }
    else if (iIconNumber > 0)
        iTypeImage = IMAGE_ICON;
    else if (sFile_EXT)
    {
        // Determine the image type based on extension.
        if (!_stricmp(sFile_EXT, "ico"))
            iTypeImage = IMAGE_ICON;
        else if (!_stricmp(sFile_EXT, "cur") || !_stricmp(sFile_EXT, "ani"))
            iTypeImage = IMAGE_CURSOR;
        else if (!_stricmp(sFile_EXT, "bmp"))
            iTypeImage = IMAGE_BITMAP;
    }

    // If iWidth or iHeight is -1, we want to keep the aspect ratio.
    if ((iWidth == -1 || iHeight == -1) && (!iWidth || !iHeight))
        iWidth = iHeight = 0;
    bool keep_aspect_ratio = (iWidth == -1 || iHeight == -1);

    HINSTANCE hinstGDI = nullptr;
    // Try to load GDI+ if it is requested.
    if (bUseGDIPlusIfAvailable && !(hinstGDI = LoadLibrary("gdiplus")))
        bUseGDIPlusIfAvailable = false;

    // If the image has not been loaded yet and we are not using GDI+...
    if (!_hBitmap && iTypeImage > -1 && !bUseGDIPlusIfAvailable)
    {
        int desired_width = keep_aspect_ratio ? 0 : iWidth;
        int desired_height = keep_aspect_ratio ? 0 : iHeight;
        // Use LoadImage() to load the image file.
        _hBitmap = reinterpret_cast<HBITMAP>(
            LoadImage(nullptr, sFileImage, iTypeImage, desired_width, desired_height,
                LR_LOADFROMFILE | LR_CREATEDIBSECTION));
        if (_hBitmap)
        {
            // If we are not adjusting dimensions, return the bitmap immediately.
            if (!keep_aspect_ratio)
                return _hBitmap;
        }
        else if (GetFileAttributes(sFileImage) == 0xFFFFFFFF)
            return nullptr; // File does not exist.
        else if (iIconNumber > 0)
        {
            // Attempt extracting the icon if loading as an image fails.
            _hBitmap = reinterpret_cast<HBITMAP>(
                ExtractIcon(GetModuleHandle(nullptr), sFileImage, iIconNumber - 1));
            if (reinterpret_cast<std::uintptr_t>(_hBitmap) < 2)
                return nullptr;
            ExtractIcon_was_used = true;
        }
    }

    IPicture* pic = nullptr;
    if (!_hBitmap)
    {
        iTypeImage = IMAGE_BITMAP;
        // If GDI+ is requested or file extension is not one of jpg/jpeg/gif, ensure GDI+ is loaded.
        if (bUseGDIPlusIfAvailable || !sFile_EXT || (_stricmp(sFile_EXT, "jpg") &&
            _stricmp(sFile_EXT, "jpeg") && _stricmp(sFile_EXT, "gif")))
            if (!hinstGDI)
                hinstGDI = LoadLibrary("gdiplus");
        if (hinstGDI)
        {
            // Define function pointers for required GDI+ functions.
            using GdiplusStartupType = int (WINAPI*)(ULONG_PTR*, LPVOID, LPVOID);
            using GdiplusShutdownType = VOID(WINAPI*)(ULONG_PTR);
            using GdipCreateBitmapFromFileType = int (WINGDIPAPI*)(LPVOID, LPVOID);
            using GdipCreateHBITMAPFromBitmapType = int (WINGDIPAPI*)(LPVOID, LPVOID, DWORD);
            using GdipDisposeImageType = int (WINAPI*)(LPVOID);
            // Dynamically retrieve function addresses from the GDI+ DLL.
            auto DynGdiplusStartup = reinterpret_cast<GdiplusStartupType>(GetProcAddress(hinstGDI, "GdiplusStartup"));
            auto DynGdiplusShutdown = reinterpret_cast<GdiplusShutdownType>(GetProcAddress(hinstGDI, "GdiplusShutdown"));
            auto DynGdipCreateBitmapFromFile = reinterpret_cast<GdipCreateBitmapFromFileType>(GetProcAddress(hinstGDI, "GdipCreateBitmapFromFile"));
            auto DynGdipCreateHBITMAPFromBitmap = reinterpret_cast<GdipCreateHBITMAPFromBitmapType>(GetProcAddress(hinstGDI, "GdipCreateHBITMAPFromBitmap"));
            auto DynGdipDisposeImage = reinterpret_cast<GdipDisposeImageType>(GetProcAddress(hinstGDI, "GdipDisposeImage"));

            ULONG_PTR token;
            Gdiplus::GdiplusStartupInput gdi_input;
            Gdiplus::GpBitmap* pgdi_bitmap = nullptr;
            // Initialize GDI+.
            if (DynGdiplusStartup && DynGdiplusStartup(&token, &gdi_input, nullptr) == Gdiplus::Ok)
            {
                WCHAR filespec_wide[MAX_PATH];
                // Convert file specification from ANSI to wide-character.
                ToWideCharFunc(sFileImage, filespec_wide, MAX_PATH);
                // Create a GDI+ Bitmap from the file.
                if (DynGdipCreateBitmapFromFile(filespec_wide, &pgdi_bitmap) == Gdiplus::Ok)
                {
                    // Convert the GDI+ Bitmap to an HBITMAP.
                    if (DynGdipCreateHBITMAPFromBitmap(pgdi_bitmap, &_hBitmap, CLR_DEFAULT) != Gdiplus::Ok)
                        _hBitmap = nullptr;
                    // Dispose the GDI+ Bitmap as it is no longer needed.
                    DynGdipDisposeImage(pgdi_bitmap);
                }
                // Shut down GDI+.
                DynGdiplusShutdown(token);
            }
            FreeLibrary(hinstGDI);
        }
        else
        {
            // Fallback option: load the file using standard file I/O.
            HANDLE hfile = CreateFile(sFileImage, GENERIC_READ, 0, nullptr, OPEN_EXISTING, 0, nullptr);
            if (hfile == INVALID_HANDLE_VALUE)
                return nullptr;
            DWORD size = GetFileSize(hfile, nullptr);
            HGLOBAL hglobal = GlobalAlloc(GMEM_MOVEABLE, size);
            if (!hglobal)
            {
                CloseHandle(hfile);
                return nullptr;
            }
            LPVOID hlocked = GlobalLock(hglobal);
            if (!hlocked)
            {
                CloseHandle(hfile);
                GlobalFree(hglobal);
                return nullptr;
            }
            // Read the file into the allocated memory.
            ReadFile(hfile, hlocked, size, &size, nullptr);
            GlobalUnlock(hglobal);
            CloseHandle(hfile);
            LPSTREAM stream;
            // Create a stream from the memory block.
            if (FAILED(CreateStreamOnHGlobal(hglobal, FALSE, &stream)) || !stream)
            {
                GlobalFree(hglobal);
                return nullptr;
            }
            // Use OleLoadPicture to load the image from the stream.
            if (FAILED(OleLoadPicture(stream, 0, FALSE, IID_IPicture, (void**)&pic)))
                pic = nullptr;
            stream->Release();
            GlobalFree(hglobal);
            if (!pic)
                return nullptr;
            // Retrieve the HBITMAP from the loaded picture.
            pic->get_Handle((OLE_HANDLE*)&_hBitmap);
            if (!_hBitmap)
            {
                pic->Release();
                return nullptr;
            }
        }
    }

    // If we need to maintain aspect ratio, adjust the target dimensions.
    if (keep_aspect_ratio)
    {
        HBITMAP hBitmap_to_Analyze;
        ICONINFO ii;
        // For bitmaps, we work directly; for icons, get the icon mask.
        if (iTypeImage == IMAGE_BITMAP)
            hBitmap_to_Analyze = _hBitmap;
        else
        {
            if (GetIconInfo((HICON)_hBitmap, &ii))
                hBitmap_to_Analyze = ii.hbmMask;
            else
            {
                DestroyIcon((HICON)_hBitmap);
                return nullptr;
            }
        }
        BITMAP bitmap;
        // Get properties of the bitmap (width, height, etc.).
        GetObject(hBitmap_to_Analyze, sizeof(BITMAP), &bitmap);
        if (iHeight == -1)
        {
            // Calculate height preserving aspect ratio.
            if (bitmap.bmWidth)
                iHeight = static_cast<int>(((double)bitmap.bmHeight / bitmap.bmWidth) * iWidth + .5);
        }
        else
        {
            // Calculate width preserving aspect ratio.
            if (bitmap.bmHeight)
                iWidth = static_cast<int>(((double)bitmap.bmWidth / bitmap.bmHeight) * iHeight + .5);
        }
        if (iTypeImage != IMAGE_BITMAP)
        {
            DeleteObject(ii.hbmColor);
            DeleteObject(ii.hbmMask);
            if (!ExtractIcon_was_used)
            {
                DestroyIcon((HICON)_hBitmap);
                return (HBITMAP)LoadImage(nullptr, sFileImage, iTypeImage, iWidth, iHeight, LR_LOADFROMFILE);
            }
        }
    }
    HBITMAP hBitmap_New;
    if (pic)
    {
        // Create a copy of the image with specified dimensions.
        hBitmap_New = (HBITMAP)CopyImage(_hBitmap, IMAGE_BITMAP, iWidth, iHeight,
            (iWidth || iHeight) ? 0 : LR_COPYRETURNORG);
        pic->Release();
    }
    else
    {
        // If no scaling was requested, simply return the original.
        if (!iWidth && !iHeight)
            return _hBitmap;
        // Create a new bitmap and delete the original afterwards.
        hBitmap_New = (HBITMAP)CopyImage(_hBitmap, iTypeImage, iWidth, iHeight,
            LR_COPYRETURNORG | LR_COPYDELETEORG);
    }
    return hBitmap_New;
}

//-------------------------------------------------------------------------------------
// getbits: Retrieve a pixel array from an HBITMAP using a DIB.
// Returns a std::vector for automatic memory management.
//-------------------------------------------------------------------------------------
std::vector<COLORREF> getbits(HBITMAP ahImage, HDC hdc, LONG& iWidth, LONG& iHeight, bool& aIs16Bit, int aMinColorDepth = 8)
{
    BITMAP bmp;
    if (GetObject(ahImage, sizeof(BITMAP), &bmp) == 0)
        return {};

    iWidth = bmp.bmWidth;
    iHeight = bmp.bmHeight;
    if (bmp.bmBitsPixel < aMinColorDepth)
        return {};

    aIs16Bit = (bmp.bmBitsPixel == 16);
    int pixelCount = iWidth * iHeight;

    std::vector<COLORREF> image_pixels(pixelCount);

    BITMAPINFO bmi;
    std::memset(&bmi, 0, sizeof(bmi));  // Initialize the structure to zero.
    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = iWidth;
    bmi.bmiHeader.biHeight = -iHeight;  // Negative for a top-down DIB.
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 32;      // Request 32-bit pixels.
    bmi.bmiHeader.biCompression = BI_RGB;

    // Create a memory device context compatible with the provided hdc.
    HDC memDC = CreateCompatibleDC(hdc);
    if (!memDC)
        return {};
    // Select the bitmap into the memory DC.
    HGDIOBJ oldObj = SelectObject(memDC, ahImage);
    int ret = GetDIBits(memDC, ahImage, 0, iHeight, image_pixels.data(), &bmi, DIB_RGB_COLORS);
    SelectObject(memDC, oldObj);
    DeleteDC(memDC);

    if (ret == 0 || ret != iHeight)
        return {};
    return image_pixels;
}

//-------------------------------------------------------------------------------------
// IconToBitmap: Convert an HICON (icon handle) to an HBITMAP.
//-------------------------------------------------------------------------------------
HBITMAP IconToBitmap(HICON ahIcon, bool aDestroyIcon)
{
    if (!ahIcon)
        return nullptr;
    HBITMAP _hBitmap = nullptr;
    HDC hdc_desktop = GetDC(HWND_DESKTOP);
    HDC hdc = CreateCompatibleDC(hdc_desktop);
    if (hdc)
    {
        ICONINFO ii;
        if (GetIconInfo(ahIcon, &ii))
        {
            BITMAP icon_bitmap;
            if (GetObject(ii.hbmColor, sizeof(BITMAP), &icon_bitmap)
                && (_hBitmap = CreateCompatibleBitmap(hdc_desktop, icon_bitmap.bmWidth, icon_bitmap.bmHeight)))
            {
                HGDIOBJ old_object = SelectObject(hdc, _hBitmap);
                if (old_object)
                {
                    RECT rect = { 0, 0, icon_bitmap.bmWidth, icon_bitmap.bmHeight };
                    HBRUSH hbrush = CreateSolidBrush(CLR_DEFAULT);
                    FillRect(hdc, &rect, hbrush);
                    DeleteObject(hbrush);
                    DrawIconEx(hdc, 0, 0, ahIcon, icon_bitmap.bmWidth, icon_bitmap.bmHeight, 0, nullptr, DI_NORMAL);
                    SelectObject(hdc, old_object);
                }
            }
            DeleteObject(ii.hbmColor);
            DeleteObject(ii.hbmMask);
        }
        DeleteDC(hdc);
    }
    ReleaseDC(HWND_DESKTOP, hdc_desktop);
    if (aDestroyIcon)
        DestroyIcon(ahIcon);
    return _hBitmap;
}

//-------------------------------------------------------------------------------------
// CheckExactMatch: Pixel‐by‐pixel matching in “exact” mode.
//-------------------------------------------------------------------------------------
bool CheckExactMatch(
    const COLORREF* screen_pixels, int screen_width, int start_index,
    const COLORREF* image_pixels, const COLORREF* image_mask,
    int image_width, int image_height, COLORREF trans_color)
{
    int image_pixel_count = image_width * image_height;
    int candidate_index = start_index;
    for (int j = 0; j < image_pixel_count; ++j)
    {
        // Check for an exact match or ignore pixels as dictated by mask/transparent color.
        if (!((screen_pixels[candidate_index] == image_pixels[j]) ||
            (image_mask && image_mask[j]) ||
            (image_pixels[j] == trans_color)))
        {
            return false;
        }
        // Move candidate_index to the next row when the end of the row is reached.
        if ((j + 1) % image_width == 0)
        {
            int row = (j + 1) / image_width;
            candidate_index = start_index + row * screen_width;
        }
        else
        {
            candidate_index++;
        }
    }
    return true;
}

//-------------------------------------------------------------------------------------
// CheckApproxMatch: Pixel matching in “approximate” mode with tolerance.
//-------------------------------------------------------------------------------------
bool CheckApproxMatch(
    const COLORREF* screen_pixels, int screen_width, int start_index,
    const COLORREF* image_pixels, const COLORREF* image_mask,
    int image_width, int image_height, int tolerance, COLORREF trans_color)
{
    int image_pixel_count = image_width * image_height;
    int candidate_index = start_index;
    for (int j = 0; j < image_pixel_count; ++j)
    {
        BYTE search_red = GetBValue(image_pixels[j]);
        BYTE search_green = GetGValue(image_pixels[j]);
        BYTE search_blue = GetRValue(image_pixels[j]);

        BYTE red_low = (tolerance > search_red) ? 0 : search_red - tolerance;
        BYTE green_low = (tolerance > search_green) ? 0 : search_green - tolerance;
        BYTE blue_low = (tolerance > search_blue) ? 0 : search_blue - tolerance;
        BYTE red_high = (tolerance > (0xFF - search_red)) ? 0xFF : search_red + tolerance;
        BYTE green_high = (tolerance > (0xFF - search_green)) ? 0xFF : search_green + tolerance;
        BYTE blue_high = (tolerance > (0xFF - search_blue)) ? 0xFF : search_blue + tolerance;

        BYTE red = GetBValue(screen_pixels[candidate_index]);
        BYTE green = GetGValue(screen_pixels[candidate_index]);
        BYTE blue = GetRValue(screen_pixels[candidate_index]);

        bool pixel_match = ((red >= red_low && red <= red_high) &&
            (green >= green_low && green <= green_high) &&
            (blue >= blue_low && blue <= blue_high)) ||
            (image_mask && image_mask[j]) ||
            (image_pixels[j] == trans_color);
        if (!pixel_match)
            return false;

        if ((j + 1) % image_width == 0)
        {
            int row = (j + 1) / image_width;
            candidate_index = start_index + row * screen_width;
        }
        else
        {
            candidate_index++;
        }
    }
    return true;
}

//-------------------------------------------------------------------------------------
// ScaleBitmap: Create a new bitmap with a new size using a DIB Section and StretchBlt.
//-------------------------------------------------------------------------------------
static HBITMAP ScaleBitmap(HBITMAP hBitmap, int newW, int newH)
{
    BITMAPINFO bmi;
    std::memset(&bmi, 0, sizeof(bmi));
    bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = newW;
    bmi.bmiHeader.biHeight = -newH;  // Negative for top-down bitmap.
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 32;
    bmi.bmiHeader.biCompression = BI_RGB;

    HDC hdcScreen = GetDC(nullptr);
    HBITMAP hBitmapScaled = CreateDIBSection(hdcScreen, &bmi, DIB_RGB_COLORS, nullptr, nullptr, 0);
    ReleaseDC(nullptr, hdcScreen);
    if (!hBitmapScaled)
        return nullptr;

    HDC hdcSrc = CreateCompatibleDC(nullptr);
    HDC hdcDest = CreateCompatibleDC(nullptr);
    if (!hdcSrc || !hdcDest) {
        if (hdcSrc) DeleteDC(hdcSrc);
        if (hdcDest) DeleteDC(hdcDest);
        DeleteObject(hBitmapScaled);
        return nullptr;
    }
    HGDIOBJ oldSrc = SelectObject(hdcSrc, hBitmap);
    HGDIOBJ oldDest = SelectObject(hdcDest, hBitmapScaled);

    BITMAP bm;
    GetObject(hBitmap, sizeof(BITMAP), &bm);
    BOOL bRet = StretchBlt(hdcDest, 0, 0, newW, newH, hdcSrc, 0, 0, bm.bmWidth, bm.bmHeight, SRCCOPY);

    SelectObject(hdcSrc, oldSrc);
    SelectObject(hdcDest, oldDest);
    DeleteDC(hdcSrc);
    DeleteDC(hdcDest);

    if (!bRet)
    {
        DeleteObject(hBitmapScaled);
        return nullptr;
    }
    return hBitmapScaled;
}

//-------------------------------------------------------------------------------------
// SearchImageWithBitmap:
// Searches within a captured screen region for a match to the provided image bitmap.
//-------------------------------------------------------------------------------------
static char* SearchImageWithBitmap(HBITMAP hImage, int imgW, int imgH,
    int iLeft, int iTop, int iRight, int iBottom,
    int iTolerance, int iTransparent)
{
    // Static buffer to hold the result string.
    static char result[64];
    result[0] = '\0';

    HDC hdc = GetDC(nullptr);
    if (!hdc) { std::snprintf(result, sizeof(result), "E<-3>"); return result; }
    HDC sdc = CreateCompatibleDC(hdc);
    if (!sdc) { ReleaseDC(nullptr, hdc); std::snprintf(result, sizeof(result), "E<-5>"); return result; }
    int search_width = iRight - iLeft + 1;
    int search_height = iBottom - iTop + 1;
    HBITMAP hBitmap_screen = CreateCompatibleBitmap(hdc, search_width, search_height);
    if (!hBitmap_screen) { DeleteDC(sdc); ReleaseDC(nullptr, hdc); std::snprintf(result, sizeof(result), "E<-6>"); return result; }
    HGDIOBJ sdc_orig_select = SelectObject(sdc, hBitmap_screen);
    if (!sdc_orig_select) { DeleteObject(hBitmap_screen); DeleteDC(sdc); ReleaseDC(nullptr, hdc); std::snprintf(result, sizeof(result), "E<-7>"); return result; }
    
    // Capture the specified screen region into our bitmap.
    if (!BitBlt(sdc, 0, 0, search_width, search_height, hdc, iLeft, iTop, SRCCOPY))
    {
        SelectObject(sdc, sdc_orig_select);
        DeleteObject(hBitmap_screen);
        DeleteDC(sdc);
        ReleaseDC(nullptr, hdc);
        std::snprintf(result, sizeof(result), "E<-8>");
        return result;
    }
    bool screen_is16 = false;
    LONG sw = 0, sh = 0;
    auto screen_pixels = getbits(hBitmap_screen, sdc, sw, sh, screen_is16);
    if (screen_pixels.empty())
    {
        SelectObject(sdc, sdc_orig_select);
        DeleteObject(hBitmap_screen);
        DeleteDC(sdc);
        ReleaseDC(nullptr, hdc);
        std::snprintf(result, sizeof(result), "E<-9>");
        return result;
    }
    bool imgIs16 = false;
    LONG iw = 0, ih = 0;
    auto image_pixels = getbits(hImage, hdc, iw, ih, imgIs16);
    if (image_pixels.empty())
    {
        SelectObject(sdc, sdc_orig_select);
        DeleteObject(hBitmap_screen);
        DeleteDC(sdc);
        ReleaseDC(nullptr, hdc);
        std::snprintf(result, sizeof(result), "E<-10>");
        return result;
    }
    int found_index = -1;
    int total_pixels = sw * sh;
    for (int i = 0; i < total_pixels; i++)
    {
        int row = i / sw;
        int col = i % sw;
        if ((ih > (sh - row)) || (iw > (sw - col)))
            continue;
        bool match = false;
        if (iTolerance == 0)
            match = CheckExactMatch(screen_pixels.data(), sw, i, image_pixels.data(), nullptr, iw, ih, (COLORREF)iTransparent);
        else
            match = CheckApproxMatch(screen_pixels.data(), sw, i, image_pixels.data(), nullptr, iw, ih, iTolerance, (COLORREF)iTransparent);
        if (match) { found_index = i; break; }
    }
    int locx = 0, locy = 0;
    if (found_index != -1)
    {
        int row = found_index / sw;
        int col = found_index % sw;
        locx = iLeft + col;
        locy = iTop + row;
        std::snprintf(result, sizeof(result), "%d|%d|%d|%d", locx, locy, iw, ih);
    }
    SelectObject(sdc, sdc_orig_select);
    DeleteObject(hBitmap_screen);
    DeleteDC(sdc);
    ReleaseDC(nullptr, hdc);
    return result;
}

//-------------------------------------------------------------------------------------
// ImageSearch (exported)
//
// Return Structure:
//   - If there is a match: the return string has the form "{<match_count>}[<match_result(s)>]"
//       For example, if one match is found, it may return: "{1}[x|y|w|h]"
//       If there are multiple matches, the individual result strings are separated by commas.
//   - If no match is found: the return string is "0[No Match Found]" (and if debugging info is enabled,
//       extra debug information will be appended in parentheses)
//   - If an error occurs: the return string is formatted as "{-error_code}[<error_message>](...debug info...)"
//       For errors, fScaleSuccess is always reported as 0.00 in the debug info.
//
// Parameters:
//   sImageFile    - String containing one or more image file paths separated by '|'.
//   iLeft, iTop, iRight, iBottom - Define the screen search region (if iRight/iBottom are not set,
//                                  the screen dimensions are used).
//   iTolerance    - Color matching tolerance (0 for exact matching).
//   iTransparent - Color to treat as transparent (ignored during matching).
//   iMultiResults  - Maximum number of results to return (0 stops after the first match).
//   iCenterPOS     - If 1, convert found coordinates to the center of the matched image.
//   iReturnDebug   - If non-zero, append additional debug information to the output string.
//   fMinScale, fMaxScale, fScaleStep - Parameters to control image scaling during searching.
//
// Returns: A C-string with the result formatted as described above.
extern "C" __declspec(dllexport) char* WINAPI ImageSearch(char* sImageFile,
    int iLeft = 0,
    int iTop = 0,
    int iRight = 0,
    int iBottom = 0,
    int iTolerance = 0,
    int iTransparent = CLR_NONE,
    int iMultiResults = 0,
    int iCenterPOS = 1,
    int iReturnDebug = 0,
    float fMinScale = 1.0f,
    float fMaxScale = 1.0f,
    float fScaleStep = 0.1f)
{
    std::string iAccumResults;  // Accumulates matching results from possibly multiple files.
    int iMatchCount = 0;        // Counter for the number of found matches.
    bool bErrorOccurred = false;
    int iErrorCode = 0;
    // Variable to record the scale factor at which a match is successfully found.
    float fScaleSuccess = 0.00f;

    // Validate and adjust input parameters.
    if (iLeft < 1)
        iLeft = 0;
    if (iLeft > 15359)
        iLeft = 15360;
    if (iTop < 1) 
        iTop = 0;
    if (iTop > 15359) 
        iTop = 15360;
    if (iRight < 1) 
        iRight = 0;
    if (iRight > 15359) 
        iRight = 15360;
    if (iBottom < 1) 
        iBottom = 0;
    if (iBottom > 15359)
        iBottom = 15360;
    if (iRight <= iLeft)
        iRight = GetSystemMetrics(SM_CXSCREEN) - 1;
    if (iBottom <= iTop)
        iBottom = GetSystemMetrics(SM_CYSCREEN) - 1;
    if (iTolerance < 1)
        iTolerance = 0;
    if (iTolerance > 254)
        iTolerance = 255;
    if (fMinScale < 0.1f || fMinScale > 1.0f)
        fMinScale = 1.0f;
    if (fMaxScale < 1.0f)
        fMaxScale = 1.0f;
    if (fMaxScale > 2.9f)
        fMaxScale = 3.0f;
    if (fScaleStep > (fMaxScale - fMinScale))
        fScaleStep = fMaxScale - fMinScale;
    if (fScaleStep <= 0.01f)
        fScaleStep = 0.01f;
    if (iTransparent < 0)
        iTransparent = CLR_NONE;
    if (iMultiResults < 1)
        iMultiResults = 0;
    if (iMultiResults > 98)
        iMultiResults = 99;
    if (iReturnDebug < 1)
        iReturnDebug = 0;
    if (iReturnDebug > 1)
        iReturnDebug = 1;

    // Disable scaling if the scale range is essentially 1.0.
    bool bScaleSearch = true;
    if (std::fabs(fMinScale - 1.0f) < 0.0001f && std::fabs(fMaxScale - 1.0f) < 0.0001f)
    {
        bScaleSearch = false;
        fMinScale = fMaxScale = 1.0f;
    }

    // Split the sImageFile string on '|' to handle multiple image files.
    std::string filesStr(sImageFile);
    size_t start = 0;
    while (true)
    {
        size_t pos = filesStr.find('|', start);
        std::string token;
        if (pos == std::string::npos)
        {
            token = filesStr.substr(start);
        }
        else
        {
            token = filesStr.substr(start, pos - start);
            start = pos + 1;
        }
        if (!token.empty())
        {
            int width = 0, height = 0, image_type = -1;
            // Load the current image from file.
            HBITMAP hOrig = LoadPicture(const_cast<char*>(token.c_str()), width, height, image_type, 0, false);
            if (!hOrig)
            {
                iErrorCode = -2;
                bErrorOccurred = true;
                break;
            }
            char sScaleResult[64] = "";
            bool foundForThisFile = false;
            if (bScaleSearch)
            {
                // Attempt multiple scales from fMinScale to fMaxScale.
                for (float scale = fMinScale; scale <= fMaxScale + 0.0001f; scale += fScaleStep)
                {
                    int newW = static_cast<int>(width * scale);
                    int newH = static_cast<int>(height * scale);
                    HBITMAP hScaled = ScaleBitmap(hOrig, newW, newH);
                    if (!hScaled)
                        continue;
                    char* searchRes = SearchImageWithBitmap(hScaled, newW, newH, iLeft, iTop, iRight, iBottom, iTolerance, iTransparent);
                    if (searchRes[0] != '\0')
                    {
                        int x, y, w, h;
                        if (sscanf_s(searchRes, "%d|%d|%d|%d", &x, &y, &w, &h) == 4)
                        {
                            if (iCenterPOS == 1)
                            {
                                // Convert top-left found coordinate to center coordinate.
                                int centerX = x + (w / 2);
                                int centerY = y + (h / 2);
                                std::snprintf(sScaleResult, sizeof(sScaleResult), "%d|%d|%d|%d", centerX, centerY, w, h);
                                fScaleSuccess = scale;
                            }
                            else
                            {
                                strcpy_s(sScaleResult, sizeof(sScaleResult), searchRes);
                                fScaleSuccess = scale;
                            }
                        }
                        foundForThisFile = true;
                        DeleteObject(hScaled);
                        break;
                    }
                    DeleteObject(hScaled);
                }
            }
            else
            {
                // If no scaling is required, search with the original image dimensions.
                char* searchRes = SearchImageWithBitmap(hOrig, width, height, iLeft, iTop, iRight, iBottom, iTolerance, iTransparent);
                if (searchRes[0] != '\0')
                {
                    int x, y, w, h;
                    if (sscanf_s(searchRes, "%d|%d|%d|%d", &x, &y, &w, &h) == 4)
                    {
                        if (iCenterPOS == 1)
                        {
                            int centerX = x + (w / 2);
                            int centerY = y + (h / 2);
                            std::snprintf(sScaleResult, sizeof(sScaleResult), "%d|%d|%d|%d", centerX, centerY, w, h);
                            fScaleSuccess = 1.0f;
                        }
                        else
                        {
                            strcpy_s(sScaleResult, sizeof(sScaleResult), searchRes);
                            fScaleSuccess = 1.0f;
                        }
                    }
                    foundForThisFile = true;
                }
            }
            DeleteObject(hOrig);

            if (foundForThisFile)
            {
                if (!iAccumResults.empty())
                    iAccumResults.append(",");
                iAccumResults.append(sScaleResult);
                ++iMatchCount;
                if (iMultiResults == 0)
                    break;
                if (iMultiResults > 0 && iMatchCount >= iMultiResults)
                    break;
            }
        }
        if (pos == std::string::npos)
            break;
    }

    std::string retStr;
    if (bErrorOccurred)
    {
        // If an error occurred, return a string in the following format:
        //   "{<negative_error_code>}[<error_message>]"
        // And if debugging is enabled, additional debug info will be appended.
        retStr = std::format("{{{}}}[{}]", iErrorCode, GetErrorMessage(iErrorCode));
    }
    else if (iMatchCount > 0)
    {
        // If one or more matches are found, return a string in the following format:
        //   "{<match_count>}[<match_result(s)>]"
        // If multiple matches exist, individual result strings are separated by commas.
        retStr = std::format("{{{}}}[{}]", iMatchCount, iAccumResults);
    }
    else
    {
        // If no match is found, return "0[]"
        retStr = "0[No Match Found]";
    }

    if (iReturnDebug != 0)
    {
        // Append debug information if requested. Debug info includes:
        //   iLeft, iTop, iRight, iBottom, iTolerance, iTransparent,
        //   iMultiResults, iCenterPOS, fMinScale, fMaxScale, fScaleStep, fScaleSuccess.
        // For errors, fScaleSuccess will be 0.00.
        retStr += std::format("({},{},{},{},{},{},{},{},{:.2f},{:.2f},{:.2f},{:.2f})",
            iLeft, iTop, iRight, iBottom, iTolerance, iTransparent, iMultiResults, iCenterPOS,
            fMinScale, fMaxScale, fScaleStep, fScaleSuccess);
    }

    // Return the resulting string from a static buffer.
    static char szAnswer[1024];
    strcpy_s(szAnswer, sizeof(szAnswer), retStr.c_str());
    szAnswer[sizeof(szAnswer) - 1] = '\0';
    return szAnswer;
}

#ifdef _DEBUG
#include <iostream>
int _tmain(int argc, _TCHAR* argv[])
{
    char* sImageFile = nullptr;
    if (argc > 1)
    {
        sImageFile = argv[1];
    }
    else
    {
        sImageFile = "C:\\Đào Văn Trong.bmp|C:\\TRONG.PRO.bmp";
    }

    // Call ImageSearch with test parameters.
    char* result = ImageSearch(sImageFile,
        0, 0, 0, 0,    // iLeft, iTop, iRight, iBottom (use full screen dimensions)
        0,            // iTolerance for exact match
        CLR_NONE,     // iTransparent (no transparent color)
        1,            // iMultiResults (first match only)
        1,            // iCenterPOS (return center coordinates)
        0,            // iReturnDebug (disable debug info)
        1.0f, 1.0f, 0.1f);  // Scaling parameters (no scaling)
    std::cout << "Result: " << result << "\n";
    return 0;
}
#endif

#ifdef _MANAGED
#pragma managed(pop)
#endif
